'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _utils = require('./utils');

var childrenId = -1;
var dataId = -1;

var attributeRegex = /\s*(?:[a-zA-Z_][a-zA-Z0-9_]*(?:\s*(?::|-)\s*[a-zA-Z_])?)+\s*/;
var parenthesizedAttributeRegex = (0, _utils.composeRegex)(/\s*(?:\{|\[|\()\s*/, attributeRegex, /\s*(?:\)|\]|\})\s*/);
var attributeAssignmentRegex = /\s*=\s*(?:{|"|')\s*/;
var spreadAttributeRegex = /\s*\{\s*\.\.\.\s*/;

var tagCloses = [{
  regex: /\s*>/
}, {
  regex: /\s*\/\s*>/
}];

var tagAttributes = [].concat(tagCloses, [{
  regex: attributeRegex
}, {
  regex: parenthesizedAttributeRegex
}, {
  regex: attributeRegex,
  alternatives: [].concat(tagCloses, [{
    regex: attributeAssignmentRegex
  }, {
    regex: spreadAttributeRegex
  }])
}, {
  regex: parenthesizedAttributeRegex,
  alternatives: [].concat(tagCloses, [{
    regex: attributeAssignmentRegex
  }, {
    regex: spreadAttributeRegex
  }])
}, {
  regex: spreadAttributeRegex
}]);

tagAttributes[2].alternatives = tagAttributes;
tagAttributes[3].alternatives = tagAttributes;

var tagMatcher = {
  regex: /</,
  alternatives: [{
    regex: /!--/
  }, {
    regex: /\s*(?:[a-zA-Z_][a-zA-Z0-9_]*(?:\s*(?::|->?|\.)\s*[a-zA-Z_])?)+\s*/,
    alternatives: tagAttributes
  }]
};

var isTagOpen = function isTagOpen(str) {
  return (0, _utils.stringMatch)(str, tagMatcher);
};

var normalizeNewLines = function normalizeNewLines(str) {
  return str.replace(/\s*\n\s*/g, ' ');
};

var escape = function escape(char, str) {
  return str.replace(new RegExp('(' + char + ')', 'g'), '\\$1');
};

var escapeQuotes = escape.bind(null, '"');

var getTagName = function getTagName(sel) {
  return sel.indexOf('u8"') === 0 ? sel.slice(0, -1).replace('u8"', '') : sel;
};

var identity = function identity(x) {
  return x;
};

var filterByType = function filterByType(type, formatter, data) {
  return data.filter(function (attribute) {
    return attribute.type === type;
  }).filter(function (obj, index, arr) {
    return arr.map(function (mapObj) {
      return mapObj.name;
    }).lastIndexOf(obj.name) === index;
  }).map(function (attribute) {
    return _extends({}, attribute, {
      value: formatter(attribute.value)
    });
  });
};

var filterAttrs = filterByType.bind(null, 'attr', identity);
var filterProps = filterByType.bind(null, 'prop', function (value) {
  return value.indexOf('emscripten::val') !== 0 ? 'emscripten::val(' + value + ')' : value;
});
var filterCallbacks = filterByType.bind(null, 'callback', identity);

var createMaps = function createMaps(maps) {
  return maps.map(function (map) {
    return map.values.length === 0 ? '' : map.id + ' {' + map.values.map(function (attr) {
      return '{' + attr.name + ', ' + attr.value + '}';
    }).join(', ') + '}';
  }).filter(function (map) {
    return map !== '';
  }).join(', ');
};

var _createVNode = void 0;

var getChildrenDeclaration = function getChildrenDeclaration(children) {
  return 'asmdom::Children {' + children.map(_createVNode).join(', ') + '}';
};

var getDataDeclaration = function getDataDeclaration(data) {
  return 'asmdom::Data (' + createMaps([{ id: 'asmdom::Attrs', values: filterAttrs(data) }, { id: 'asmdom::Props', values: filterProps(data) }, { id: 'asmdom::Callbacks', values: filterCallbacks(data) }]) + ')';
};

var splitArray = function splitArray(arr, predicate) {
  var result = [];
  var slicedArray = arr;
  while (slicedArray.length !== 0) {
    if (predicate(slicedArray[0])) {
      result.push(slicedArray[0]);
      slicedArray = slicedArray.slice(1, slicedArray.length);
    } else {
      var vectorIndex = slicedArray.findIndex(predicate);
      if (vectorIndex !== -1) {
        result.push(slicedArray.slice(0, vectorIndex));
        slicedArray = slicedArray.slice(vectorIndex);
      } else {
        result.push(slicedArray);
        slicedArray = [];
      }
    }
  }
  return result;
};

var aggregateStrings = function aggregateStrings(vnodes) {
  if (vnodes.length === 1) return vnodes[0];
  return {
    type: 'string',
    aggregated: true,
    value: vnodes.map(function (vnode, index) {
      var value = vnode.value;


      if (vnode.type === 'CPXText') {
        if (index === 0) {
          value = value.replace(/^\s+/, '');
        } else if (index === vnodes.length - 1) {
          value = value.replace(/\s+$/, '');
        }
        return 'u8"' + value + '"';
      } else if (vnode.type === 'string') {
        return 'std::string(' + value.trim() + ')';
      }
      return '';
    }).join(' + ')
  };
};

var aggregateNodes = function aggregateNodes(vnodes) {
  var computedVnodes = [];
  var strings = [];
  for (var i = 0; i < vnodes.length; i += 1) {
    var child = vnodes[i];
    if (child.type === 'CPXText' || child.type === 'string') {
      strings.push(child);
    } else {
      if (strings.length !== 0) {
        computedVnodes.push(aggregateStrings(strings));
        strings = [];
      }
      computedVnodes.push(child);
    }
  }
  if (strings.length !== 0) {
    computedVnodes.push(aggregateStrings(strings));
  }
  return computedVnodes;
};

_createVNode = function createVNode(vnode) {
  var compiledVNode = void 0;
  if (vnode.type === 'VNode') return vnode.value;
  if (vnode.type === 'string') return 'asmdom::h(' + vnode.value.trim() + ', true)';
  if (vnode.type === 'CPXText') return 'asmdom::h(u8"' + vnode.value.trim() + '", true)';
  if (vnode.type === 'CPXComment') return 'asmdom::h(u8"!", std::string(u8"' + escapeQuotes(vnode.value) + '"))';
  if (vnode.type === 'CPXElement') {
    compiledVNode = 'asmdom::h(' + vnode.sel;

    if (vnode.data !== undefined) {
      if (vnode.data.length === 1 && vnode.data[0].type === 'spread') {
        compiledVNode += ', ' + vnode.data[0].value;
      } else if (vnode.data.length > 0) {
        if (vnode.data.find(function (x) {
          return x.type === 'spread';
        }) !== undefined) {
          // eslint-disable-next-line
          var resultVar = '_asmdom_data_concat_' + ++dataId;
          var dataGroups = splitArray(vnode.data, function (x) {
            return x.type === 'spread';
          }).map(function (group) {
            return {
              // eslint-disable-next-line
              identifier: '_asmdom_data_concat_' + ++dataId,
              value: group
            };
          });
          var identifiers = dataGroups.map(function (x) {
            return x.identifier;
          }).reverse();
          compiledVNode += ', [&]() -> asmdom::Data {';
          compiledVNode += 'asmdom::Data ' + resultVar + ';';
          compiledVNode += dataGroups.map(function (_ref) {
            var identifier = _ref.identifier,
                value = _ref.value;
            return 'asmdom::Data ' + identifier + ' = ' + (value.type === 'spread' ? value.value : getDataDeclaration(value)) + ';';
          }).join('');
          compiledVNode += identifiers.map(function (x) {
            var result = resultVar + '.attrs.insert(' + x + '.attrs.begin(), ' + x + '.attrs.end());';
            result += resultVar + '.props.insert(' + x + '.props.begin(), ' + x + '.props.end());';
            result += resultVar + '.callbacks.insert(' + x + '.callbacks.begin(), ' + x + '.callbacks.end());';
            return result;
          }).join('');
          compiledVNode += 'return ' + resultVar + ';}()';
        } else {
          compiledVNode += ', ' + getDataDeclaration(vnode.data);
        }
      }
    }

    if (vnode.children !== undefined) {
      var children = aggregateNodes(vnode.children.filter(function (child) {
        return child.type !== 'comment';
      }));

      if (children.length === 1) {
        compiledVNode += ', ';
        if (children[0].type === 'CPXText') {
          compiledVNode += 'std::string(u8"' + children[0].value.trim() + '")';
        } else if (children[0].type === 'string') {
          compiledVNode += children[0].aggregated !== true ? 'std::string(' + children[0].value.trim() + ')' : children[0].value.trim();
        } else if (children[0].type === 'Children') {
          compiledVNode += children[0].value;
        } else {
          compiledVNode += _createVNode(children[0]);
        }
      } else if (children.length > 1) {
        if (children.find(function (child) {
          return child.type === 'Children';
        }) !== undefined) {
          // eslint-disable-next-line
          var _resultVar = '_asmdom_ch_concat_' + ++childrenId;
          var childrenGroups = splitArray(children, function (x) {
            return x.type === 'Children';
          }).map(function (group) {
            return {
              // eslint-disable-next-line
              identifier: '_asmdom_ch_concat_' + ++childrenId,
              value: group
            };
          });
          var _identifiers = childrenGroups.map(function (x) {
            return x.identifier;
          });
          compiledVNode += ', [&]() -> asmdom::Children {';
          compiledVNode += 'asmdom::Children ' + _resultVar + ';';
          compiledVNode += childrenGroups.map(function (_ref2) {
            var identifier = _ref2.identifier,
                value = _ref2.value;
            return 'asmdom::Children ' + identifier + ' = ' + (value.type === 'Children' ? value.value : getChildrenDeclaration(value)) + ';';
          }).join('');
          compiledVNode += _resultVar + '.reserve(' + _identifiers.map(function (x) {
            return x + '.size()';
          }).join(' + ') + ');';
          compiledVNode += _identifiers.map(function (x) {
            return _resultVar + '.insert(' + _resultVar + '.end(), ' + x + '.begin(), ' + x + '.end());';
          }).join('');
          compiledVNode += 'return ' + _resultVar + ';}()';
        } else {
          compiledVNode += ', ' + getChildrenDeclaration(children);
        }
      }
    }

    compiledVNode += ')';
  }
  return compiledVNode;
};

exports['default'] = {
  createVNode: _createVNode,
  normalizeNewLines: normalizeNewLines,
  escapeQuotes: escapeQuotes,
  getTagName: getTagName,
  isTagOpen: isTagOpen
};